==> getml-1.0.0-py3-none-any.whl: getml/feature_learning/relboost.py <==
# Copyright 2021 The SQLNet Company GmbH

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""
Feature learning based on Gradient Boosting.
"""

from dataclasses import dataclass
from typing import Union

from .fastprop import FastProp
from .feature_learner import _FeatureLearner
from .validation import _validate_relboost_parameters

# --------------------------------------------------------------------


@dataclass(repr=False)
class Relboost(_FeatureLearner):
    """Feature learning based on Gradient Boosting.

    :class:`~getml.feature_learning.Relboost` automates feature learning
    for relational data and time series. It is based on a
    generalization of the XGBoost algorithm to relational data, hence
    the name.

    For more information on the underlying feature learning
    algorithm, check out :ref:`feature_learning_algorithms_relboost`.

    Args:
        allow_null_weights (bool, optional):
            Whether you want to allow
            :class:`~getml.feature_learning.Relboost` to set weights to
            NULL.

        delta_t (float, optional):
            Frequency with which lag variables will be explored in a
            time series setting. When set to 0.0, there will be no lag
            variables.

            For more information, please refer to
            :ref:`data_model_time_series`. Range: [0, :math:`\\infty`]

        gamma (float, optional):
            During the training of Relboost, which is based on
            gradient tree boosting, this value serves as the minimum
            improvement in terms of the `loss_function` required for a
            split of the tree to be applied. Larger `gamma` will lead
            to fewer partitions of the tree and a more conservative
            algorithm. Range: [0, :math:`\\infty`]

        loss_function (:class:`~getml.feature_learning.loss_functions`, optional):
            Objective function used by the feature learning algorithm
            to optimize your features. For regression problems use
            :class:`~getml.feature_learning.loss_functions.SquareLoss` and for
            classification problems use
            :class:`~getml.feature_learning.loss_functions.CrossEntropyLoss`.

        max_depth (int, optional):
            Maximum depth of the trees generated during the gradient
            tree boosting. Deeper trees will result in more complex
            models and increase the risk of overfitting. Range: [0,
            :math:`\\infty`]

        min_df (int, optional):
            Only relevant for columns with role :const:`~getml.data.roles.text`.
            The minimum
            number of fields (i.e. rows) in :const:`~getml.data.roles.text` column a
            given word is required to appear in to be included in the bag of words.
            Range: [1, :math:`\\infty`]

        min_num_samples (int, optional):
            Determines the minimum number of samples a subcondition
            should apply to in order for it to be considered. Higher
            values lead to less complex statements and less danger of
            overfitting. Range: [1, :math:`\\infty`]

        num_features (int, optional):
            Number of features generated by the feature learning
            algorithm. Range: [1, :math:`\\infty`]

        num_subfeatures (int, optional):
            The number of subfeatures you would like to extract in a
            subensemble (for snowflake data model only). See
            :ref:`data_model_snowflake_schema` for more
            information. Range: [1, :math:`\\infty`]

        num_threads (int, optional):
            Number of threads used by the feature learning algorithm. If set to
            zero or a negative value, the number of threads will be
            determined automatically by the getML engine. Range:
            [:math:`0`, :math:`\\infty`]

        propositionalization (:class:`~getml.feature_learning.FastProp`, optional):
            The feature learner used for joins which are flagged to be
            propositionalized (by setting a join's `relationship` parameter to
            :const:`getml.data.relationship.propositionalization`)

        reg_lambda (float, optional):
            L2 regularization on the weights in the gradient boosting
            routine. This is one of the most important hyperparameters
            in the :class:`~getml.feature_learning.Relboost` as it allows
            for the most direct regularization. Larger values will
            make the resulting model more conservative. Range: [0,
            :math:`\\infty`]

        sampling_factor (float, optional):
            Relboost uses a bootstrapping procedure (sampling with
            replacement) to train each of the features. The sampling
            factor is proportional to the share of the samples
            randomly drawn from the population table every time
            Relboost generates a new feature. A lower sampling factor
            (but still greater than 0.0), will lead to less danger of
            overfitting, less complex statements and faster
            training. When set to 1.0, roughly 20,000 samples are drawn
            from the population table. If the population table
            contains less than 20,000 samples, it will use standard
            bagging. When set to 0.0, there will be no sampling at
            all. Range: [0, :math:`\\infty`]

        seed (Union[int,None], optional):
            Seed used for the random number generator that underlies
            the sampling procedure to make the calculation
            reproducible. Internally, a `seed` of None will be mapped to
            5543. Range: [0, :math:`\\infty`]

        shrinkage (float, optional):
            Since Relboost works using a gradient-boosting-like
            algorithm, `shrinkage` (or learning rate) scales down the
            weights and thus the impact of each new tree. This gives
            more room for future ones to improve the overall
            performance of the model in this greedy algorithm. It must
            be between 0.0 and 1.0 with higher values leading to more
            danger of overfitting. Range: [0, 1]

        silent (bool, optional):
            Controls the logging during training.

        vocab_size (int, optional):
            Determines the maximum number
            of words that are extracted in total from :const:`getml.data.roles.text`
            columns. This can be interpreted as the maximum size of the bag of words.
            Range: [0, :math:`\\infty`]
    """

    # ----------------------------------------------------------------

    allow_null_weights: bool = False
    delta_t: float = 0.0
    gamma: float = 0.0
    loss_function: Union[str, None] = None
    max_depth: int = 3
    min_df: int = 30
    min_num_samples: int = 1
    num_features: int = 100
    num_subfeatures: int = 100
    num_threads: int = 0
    propositionalization: FastProp = FastProp()
    reg_lambda: float = 0.0
    sampling_factor: float = 1.0
    seed: int = 5543
    shrinkage: float = 0.1
    silent: bool = True
    vocab_size: int = 500

    # ------------------------------------------------------------

    def validate(self, params=None):
        """Checks both the types and the values of all instance
        variables and raises an exception if something is off.

        Args:
            params (dict, optional): A dictionary containing
                the parameters to validate. If not is passed,
                the own parameters will be validated.

        """

        # ------------------------------------------------------------

        if params is None:
            params = self.__dict__
        else:
            params = {**self.__dict__, **params}

        # ------------------------------------------------------------

        if not isinstance(params, dict):
            raise ValueError("params must be None or a dictionary!")

        # ------------------------------------------------------------

        for kkey in params:
            if kkey not in type(self)._supported_params:
                raise KeyError(
                    f"Instance variable '{kkey}' is not supported in {self.type}."
                )

        # ------------------------------------------------------------

        if not isinstance(params["silent"], bool):
            raise TypeError("'silent' must be of type bool")

        # ------------------------------------------------------------

        _validate_relboost_parameters(**params)

        # ------------------------------------------------------------

==> getml-1.5.1-py3-none-any.whl: getml/feature_learning/relboost.py <==
# Copyright 2025 Code17 GmbH
#
# This file is licensed under the Elastic License 2.0 (ELv2).
# Refer to the LICENSE.txt file in the root of the repository
# for details.
#


"""
Feature learning based on Gradient Boosting.
"""

from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Union

from .fastprop import FastProp
from .feature_learner import _FeatureLearner
from .loss_functions import CrossEntropyLossType, SquareLossType
from .validation import _validate_relboost_parameters

# --------------------------------------------------------------------


@dataclass(repr=False)
class Relboost(_FeatureLearner):
    """
       Feature learning based on Gradient Boosting.

    [`Relboost`][getml.feature_learning.Relboost] automates feature learning
    for relational data and time series. It is based on a
    generalization of the XGBoost algorithm to relational data, hence
    the name.

    For more information on the underlying feature learning
    algorithm, check out the User Guide: [Relboost][feature-engineering-algorithms-relboost].

    enterprise-adm: Enterprise edition
        This feature is exclusive to the Enterprise edition and is not available in the Community edition. Discover the [benefits of the Enterprise edition][enterprise-benefits] and [compare their features][enterprise-feature-list].

        For licensing information and technical support, please [contact us][contact-page].

    Args:
        allow_null_weights:
            Whether you want to allow
            [`Relboost`][getml.feature_learning.Relboost] to set weights to
            NULL.

        delta_t:
            Frequency with which lag variables will be explored in a
            time series setting. When set to 0.0, there will be no lag
            variables.

            For more information, please refer to
            [Time Series][data-model-time-series] in the User Guide. Range: [0, ∞]

        gamma:
            During the training of Relboost, which is based on
            gradient tree boosting, this value serves as the minimum
            improvement in terms of the `loss_function` required for a
            split of the tree to be applied. Larger `gamma` will lead
            to fewer partitions of the tree and a more conservative
            algorithm. Range: [0, ∞]

        loss_function:
            Objective function used by the feature learning algorithm
            to optimize your features. For regression problems use
            [`SquareLoss`][getml.feature_learning.loss_functions.SQUARELOSS] and for
            classification problems use
            [`CrossEntropyLoss`][getml.feature_learning.loss_functions.CROSSENTROPYLOSS].

        max_depth:
            Maximum depth of the trees generated during the gradient
            tree boosting. Deeper trees will result in more complex
            models and increase the risk of overfitting. Range: [0,
            ∞]

        min_df:
            Only relevant for columns with role [`text`][getml.data.roles.text].
            The minimum
            number of fields (i.e. rows) in [`text`][getml.data.roles.text] column a
            given word is required to appear in to be included in the bag of words.
            Range: [1, ∞]

        min_num_samples:
            Determines the minimum number of samples a subcondition
            should apply to in order for it to be considered. Higher
            values lead to less complex statements and less danger of
            overfitting. Range: [1, ∞]

        num_features:
            Number of features generated by the feature learning
            algorithm. Range: [1, ∞]

        num_subfeatures:
            The number of subfeatures you would like to extract in a
            subensemble (for snowflake data model only). See
            [Snowflake Schema][data-model-snowflake-schema] for more
            information. Range: [1, ∞]

        num_threads:
            Number of threads used by the feature learning algorithm. If set to
            zero or a negative value, the number of threads will be
            determined automatically by the getML Engine. Range:
            [0, ∞]

        propositionalization:
            The feature learner used for joins which are flagged to be
            propositionalized (by setting a join's `relationship` parameter to
            [`propositionalization`][getml.data.relationship.propositionalization])

        reg_lambda:
            L2 regularization on the weights in the gradient boosting
            routine. This is one of the most important hyperparameters
            in the [`Relboost`][getml.feature_learning.Relboost] as it allows
            for the most direct regularization. Larger values will
            make the resulting model more conservative. Range: [0,
            ∞]

        sampling_factor:
            Relboost uses a bootstrapping procedure (sampling with
            replacement) to train each of the features. The sampling
            factor is proportional to the share of the samples
            randomly drawn from the population table every time
            Relboost generates a new feature. A lower sampling factor
            (but still greater than 0.0), will lead to less danger of
            overfitting, less complex statements and faster
            training. When set to 1.0, roughly 20,000 samples are drawn
            from the population table. If the population table
            contains less than 20,000 samples, it will use standard
            bagging. When set to 0.0, there will be no sampling at
            all. Range: [0, ∞]

        seed:
            Seed used for the random number generator that underlies
            the sampling procedure to make the calculation
            reproducible. Internally, a `seed` of None will be mapped to
            5543. Range: [0, ∞]

        shrinkage:
            Since Relboost works using a gradient-boosting-like
            algorithm, `shrinkage` (or learning rate) scales down the
            weights and thus the impact of each new tree. This gives
            more room for future ones to improve the overall
            performance of the model in this greedy algorithm. It must
            be between 0.0 and 1.0 with higher values leading to a higher
            danger of overfitting. Range: [0, 1]

        silent:
            Controls the logging during training.

        vocab_size:
            Determines the maximum number
            of words that are extracted in total from [`text`][getml.data.roles.text]
            columns. This can be interpreted as the maximum size of the bag of words.
            Range: [0, ∞]

    """

    # ----------------------------------------------------------------

    allow_null_weights: bool = False
    delta_t: float = 0.0
    gamma: float = 0.0
    loss_function: Optional[Union[CrossEntropyLossType, SquareLossType]] = None
    max_depth: int = 3
    min_df: int = 30
    min_num_samples: int = 1
    num_features: int = 100
    num_subfeatures: int = 100
    num_threads: int = 0
    propositionalization: FastProp = field(default_factory=FastProp)
    reg_lambda: float = 0.0
    sampling_factor: float = 1.0
    seed: int = 5543
    shrinkage: float = 0.1
    silent: bool = True
    vocab_size: int = 500

    # ------------------------------------------------------------

    def validate(self, params: Optional[Dict[str, Any]] = None) -> None:
        """
        Checks both the types and the values of all instance
        variables and raises an exception if something is off.

        Args:
            params: A dictionary containing
                the parameters to validate. If not is passed,
                the own parameters will be validated.

        """

        # ------------------------------------------------------------

        if params is None:
            params = self.__dict__
        else:
            params = {**self.__dict__, **params}

        # ------------------------------------------------------------

        if not isinstance(params, dict):
            raise ValueError("params must be None or a dictionary!")

        # ------------------------------------------------------------

        for kkey in params:
            if kkey not in type(self)._supported_params:
                raise KeyError(
                    f"Instance variable '{kkey}' is not supported in {self.type}."
                )

        # ------------------------------------------------------------

        if not isinstance(params["silent"], bool):
            raise TypeError("'silent' must be of type bool")

        # ------------------------------------------------------------

        _validate_relboost_parameters(**params)

        # ------------------------------------------------------------
